/*
 * Copyright (c) 2018 by Intinor AB. All rights reserved.
 *
 * This file is part of "libRaptorQ".
 *
 * libRaptorQ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * libRaptorQ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * and a copy of the GNU Lesser General Public License
 * along with libRaptorQ.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <immintrin.h>
#include "RaptorQ/v1/gf256.hpp"
#include "RaptorQ/v1/multiplication.hpp"

namespace RaptorQ__v1 {
namespace Impl {
namespace Matrix_Row_SIMD {
    void add_avx2(uint8_t *dest, uint8_t *src, int bytes)
    {
        uint8_t *src_ptr, *dest_ptr;
        __m256i s, d;

        assert(bytes % 32 == 0);
        assert(bytes >= 32);

        src_ptr = src;
        dest_ptr = dest;

        for (int i = 0; i < bytes; i += 32) {
            s = _mm256_load_si256((__m256i *) (&src_ptr[i]));
            d = _mm256_load_si256((__m256i *) (&dest_ptr[i]));
            d = _mm256_xor_si256(d, s);
            _mm256_store_si256((__m256i *) (&dest_ptr[i]), d);
        }
    }

    void div_avx2(uint8_t *data, uint8_t num, int bytes)
    {
        uint8_t *high, *low, *data_ptr;
        __m256i  mask, t, r, v, m_table_high, m_table_low;

        assert(bytes % 32 == 0 && bytes >= 32);

        // Inverse num (div is multiply with 1/num)
        num = oct_exp[255 - oct_log[num - 1]];

        high = gf256_high_ptr + (num << 4) * 2;
        low  = gf256_low_ptr + (num << 4) * 2;

        mask = _mm256_set1_epi8 (0x0f);
        m_table_high = _mm256_loadu_si256 ((__m256i *)(high));
        m_table_low  = _mm256_loadu_si256 ((__m256i *)(low));

        data_ptr = data;
        for (int i = 0; i < bytes; i += 32) {
            v = _mm256_load_si256 ((__m256i *)(&data_ptr[i]));
            t = _mm256_and_si256 (mask, v);
            r = _mm256_shuffle_epi8 (m_table_low, t);
            v = _mm256_srli_epi64 (v, 4);
            t = _mm256_and_si256 (mask, v);
            r = _mm256_xor_si256 (r, _mm256_shuffle_epi8 (m_table_high, t));
            _mm256_store_si256 ((__m256i *)(&data_ptr[i]), r);
        }
    }

    void multiply_and_add_avx2(uint8_t *dest, uint8_t *src, uint8_t num,
                                                                    int bytes)
    {
        uint8_t *high, *low, *src_ptr, *dest_ptr;
        __m256i  mask, t, r, v, m_table_high, m_table_low;

        assert(bytes % 32 == 0 && bytes >= 32);

        high = gf256_high_ptr + (num << 4) * 2;
        low  = gf256_low_ptr + (num << 4) * 2;

        mask = _mm256_set1_epi8 (0x0f);
        m_table_high = _mm256_loadu_si256 ((__m256i *)(high));
        m_table_low  = _mm256_loadu_si256 ((__m256i *)(low));

        src_ptr = src;
        dest_ptr = dest;
        for (int i = 0; i < bytes; i += 32) {
            v = _mm256_load_si256 ((__m256i *)(&src_ptr[i]));
            t = _mm256_and_si256 (mask, v);
            r = _mm256_shuffle_epi8 (m_table_low, t);
            v = _mm256_srli_epi64 (v, 4);
            t = _mm256_and_si256 (mask, v);
            r = _mm256_xor_si256 (r, _mm256_shuffle_epi8 (m_table_high, t));
            v = _mm256_load_si256 ((__m256i *)(&dest_ptr[i]));
            r = _mm256_xor_si256 (r, v);
            _mm256_store_si256 ((__m256i *)(&dest_ptr[i]), r);
        }
    }

    uint32_t non_zero_avx2(uint8_t *data, uint32_t start, uint32_t stop)
    {
        uint8_t *data_ptr;
        uint32_t non_zero_count = 0;
        const __m256i zero32 = _mm256_setzero_si256();

        // SIMD values
        uint32_t limit = ((stop - start) / 32) * 32;

        __m256i v;
        data_ptr = data + start;
        non_zero_count += limit;
        for (uint32_t i = 0; i < limit; i += 32) {
            v = _mm256_loadu_si256((const __m256i*)&data_ptr[i]);
            v = _mm256_cmpeq_epi8(v, zero32);
            uint32_t k = _mm256_movemask_epi8(v);
            non_zero_count -= _mm_popcnt_u32(k);
        }

        // Post SIMD values
        for (uint32_t j = start + limit; j < stop; j++)
            non_zero_count += (data[j] != 0);

        return non_zero_count;
    }

    uint32_t first_non_zero_avx2(uint8_t *data, uint32_t start, uint32_t stop)
    {
        uint8_t *data_ptr;
        __m256i zero32, cmp;

        // SIMD values
        uint32_t limit = ((stop - start) / 32) * 32;
        zero32 = _mm256_setzero_si256();
        data_ptr = data + start;
        for (uint32_t i = 0; i < limit; i += 32) {
            cmp = _mm256_loadu_si256((const __m256i*)&data_ptr[i]);
            cmp = _mm256_cmpeq_epi8(cmp, zero32);
            uint32_t mask = _mm256_movemask_epi8(cmp);
            if(mask != 0xffffffff) {
                for (uint32_t j = i; j < i + 32; j++) {
                    if (data_ptr[j] != 0) {
                        return start + j;
                    }
                }
            }
        }

        // Post SIMD values
        for (uint32_t j = start + limit; j < stop; j++) {
            if (data[j] != 0) {
                return j;
            }
        }
        assert(0);
        return 0;
    }
} // namespace Matrix_Row_SIMD
} // namespace Impl
} // namespace RaptorQ__v1
